import os
import sys
import select
import logging
import traceback
import threading

from socket import SO_RCVBUF
from socket import SOL_SOCKET
from socket import error as SocketError
from pyroute2.common import DEFAULT_RCVBUF
try:
    import Queue
except ImportError:
    import queue as Queue


class IOLoop(threading.Thread):
    def __init__(self):
        threading.Thread.__init__(self, name="IOLoop")
        self.fd, self.control = os.pipe()
        self._stop_flag = False
        self.fds = {}
        self.rcvbuf = {}
        self.poll = select.epoll()
        self.register(self.fd, lambda f, event: os.read(f, 1))
        self.buffers = Queue.Queue()
        self._dequeue_thread = threading.Thread(target=self._dequeue,
                                                name='Buffers queue')
        if hasattr(sys, 'ps1'):
            self.setDaemon(True)
            self._dequeue_thread.setDaemon(True)

    def _dequeue(self):
        while True:
            cb, fd, data, argv, kwarg = self.buffers.get()

            if self._stop_flag:
                break

            try:
                cb(fd, data, *argv, **kwarg)
            except:
                logging.warning(traceback.format_exc())

    def shutdown(self):
        self._stop_flag = True
        self.reload()
        self.buffers.put((None, None, None, None, None))
        self._dequeue_thread.join()
        # wait the main loop to exit
        self.join()
        # close own file descriptors
        os.close(self.control)
        os.close(self.fd)
        self.poll.close()

    def reload(self):
        os.write(self.control, b's')

    def register(self, fd, cb, argv=[], kwarg={},
                 defer=False, proxy=None):

        # get FD number and set rcvbuf
        if isinstance(fd, int):
            fdno = fd
            self.rcvbuf[fdno] = DEFAULT_RCVBUF
        else:
            fdno = fd.fileno()
            # SO_RCVBUF // 2: see man 7 socket
            self.rcvbuf[fdno] = fd.getsockopt(SOL_SOCKET,
                                              SO_RCVBUF) // 2
        assert fdno != self.control

        if defer:
            # speed optimization
            if proxy is not None:
                # use proxy to get the next message
                def wrap(fd, event, *argv, **kwarg):
                    try:
                        data = proxy.get(fd, self.rcvbuf[fdno])
                    except OSError:
                        # EOF
                        data = ''
                    self.buffers.put((cb, fd, data, argv, kwarg))
            else:
                # or get it directly from the socket
                def wrap(fd, event, *argv, **kwarg):
                    try:
                        data = fd.recv(self.rcvbuf[fdno])
                    except OSError:
                        # EOF
                        data = ''
                    self.buffers.put((cb, fd, data, argv, kwarg))
            self.fds[fdno] = [wrap, fd, argv, kwarg]
        else:
            self.fds[fdno] = [cb, fd, argv, kwarg]

        self.poll.register(fdno, select.EPOLLIN)
        self.reload()

    def unregister(self, fd):
        try:
            fdno = fd.fileno()
        except SocketError:
            return False

        assert fdno != self.control

        if fdno in self.fds:
            del self.fds[fdno]
            self.poll.unregister(fdno)
            self.reload()
            return True

        return False

    def run(self):
        self._dequeue_thread.start()
        while True:
            try:
                fds = self.poll.poll()
            except IOError:
                continue

            if self._stop_flag:
                break

            for (fdno, event) in fds:
                try:
                    rc = self.fds[fdno]
                    rc[0](rc[1], event, *rc[2], **rc[3])
                except KeyError:
                    pass
                except:
                    logging.warning(traceback.format_exc())
